//** Modified-Light version of Drawable-Mesh for TP**// 




/*
** Drawable_mesh.h
*/

#ifndef   	DRAWABLE_MESH_H_
# define   	DRAWABLE_MESH_H_

#include <stdlib.h>
#include <iostream>
#include <cmath>
#include <string>
#include <errno.h>
#include <stdio.h>

#include <Mesh_object.h>

#include <vector>


/// Mesh class for fast drawing method
/** Container class for an optimized mesh drawing

    Use same connectivity for color and texture than geometry
*/
class Drawable_mesh
{
 public:

  //************************************************//
  //************************************************//
  //CONSTRUCTORS
  //************************************************//
  //************************************************//

  /// classical constructor
  Drawable_mesh();
  /// Copy constructor
  Drawable_mesh(const Drawable_mesh& );
  /// classical destructor
  ~Drawable_mesh();
  /// Set all size to 0
  int destroy();


  //************************************************//
  //************************************************//
  //BUILD DRAWABLE MESH
  //************************************************//
  //************************************************//

  /// Build a drawable mesh given a complete Mesh_object
  int build_drawable(const Mesh_object& _mesh);


  //************************************************//
  //************************************************//
  //Information and access
  //************************************************//
  //************************************************//

  /// get the number of triangles
  int triangle_number() const;
  /// get the number of quads
  int quad_number() const;
  /// get the number of polygons of size>4
  int polygon_number() const;
  /// get the size of the polygon k_polygon
  int polygon_size(int k_polygon) const;

  /// is normal active
  int is_normal() const;
  /// is texture active
  int is_texture() const;
  /// is color active
  int is_color() const;


  /// get the pointer on the point_set
  double* point_set_pointer();
  /// get the pointer on the normal
  double* normal_pointer();
  /// get the pointer on the texture
  double* texture_pointer();
  /// get the pointer on the color
  double* color_pointer();
  /// get the pointer on the triangle connectivity
  int* connectivity_triangles_pointer();
  /// get the pointer on the quad connectivity
  int* connectivity_quads_pointer();
  /// get the pointer on the other polygons connectivity
  int* connectivity_polygons_size_pointer();
  /// get the pointer on the other polygons connectivity
  int* connectivity_polygons_pointer();

  /// get the coordinates of a vertex
  double get_point_set(int k_vertex,int k_dim) const;
  /// get the triangle connectivity of the triangle k_triangle
  double get_connectivity_triangles(int k_triangle,int k_dim) const;
  /// get the quad connectivity of the quad k_quad
  double get_connectivity_quads(int k_quad,int k_dim) const;

  /// set the texture number
  int set_texture_number(int _texture_number);
  /// get the texture number
  int get_texture_number() const;

 private:


  /// the geometry vector
  std::vector <double> point_set;
  std::vector <double> normal;
  std::vector <double> color;
  std::vector <double> texture;

  /// the triangle connectivity
  std::vector <int> connectivity_triangles;

  /// the quad connectivity
  std::vector <int> connectivity_quads;

  /// others polygons
  std::vector <int> size_other_polygons;
  std::vector <int> connectivity_polygons;

  int texture_number;
};

#endif
