//** Modified-Light version of Segment for TP**// 


#ifndef   	SEGMENT_H_
# define   	SEGMENT_H_

#include <V_3D.h>

/// Manipulation class for a Segment
/** Calculation for embedeed Segment. */

class Segment
{
public:

  //*********************************************//
  // CONSTRUCTORS
  //*********************************************//

  /// classical constructor
  Segment();
  /// direct constructor
  Segment(const V_3D& start_point,const V_3D& end_point);
  /// Copy constructor
  Segment(const Segment& s);

  /// destructor
  ~Segment();



  //*********************************************//
  // Manipulation
  //*********************************************//

  /// Set the values
  int set(const V_3D x0,const V_3D x1);
  /// get the first point
  const V_3D& get_start() const;
  /// get the last point
  const V_3D& get_last() const;

  /// get the vector (x0,x1) / ||(x0,x1)||
  V_3D unit_vector() const;

  /// inverse the segment (first <-> last)
  int flip();



  /// empty the values
  int destroy();


  //*********************************************//
  // Informations
  //*********************************************//

  /// return the length of the segment
  double length() const;

  /// return the closest point of x on the segment
  V_3D closest_point(const V_3D& x) const;

  /// return the distance of point x to the closest point on the segment
  double distance_to_point(const V_3D& x) const;

  /// return the segment of minimal length between two segments
  /** 
      There is only 2 posibilities
      1. The closest point links the extremities
      2. The closest point is the perpendicular to both segments

      (The first point of the returned segment belongs to the current segment)
  */
  Segment closest_to_segment(const Segment& s) const;

  /// get the minimal distance between the two segments 
  /** based on the function closest_point_in_segment */
  double distance_to_segment(const Segment& s) const;

  /// Intersection with the plane of equation <n,x-x0>=0
  /** The plane is defined by its normal n, and a point x0, so its equation is <n,x-x0>=0
      4 possibilities: _ no intersections (type 0)
                       _ plane intersect on one interior points (type 1)
		       _ plane intersect on a vertex (type 2)
		       _ plane contains the segment (type 3)
      
      Solving the system <n,x-x0>=0 & A+t*u=x leads to t=<n,x0-A>/<n,u>
   */
  V_3D plane_intersection(const V_3D& n,const V_3D& x0,int *type) const;


  //*********************************************//
  // Operators
  //*********************************************//

  /// direct access operator
  V_3D  operator()(int index) const;
  /// direct access operator
  V_3D& operator()(int index);
  /// direct access operator
  V_3D  operator[](int index) const;
  /// direct access operator
  V_3D  operator[](int index);

  /// Print operator
  friend ostream& operator << (ostream& flux, const Segment& s);
  
private:

  /// First and last point
  V_3D x0,x1;

};


#endif 	    /* !SEGMENT_H_ */
