//** Modified-Light version of Drawable-Mesh for TP**// 




#include <Drawable_mesh.h>

Drawable_mesh::Drawable_mesh()
{texture_number=-1;}
Drawable_mesh::Drawable_mesh(const Drawable_mesh& _draw)
{
  texture_number         = _draw.texture_number;
  point_set              = _draw.point_set;
  color                  = _draw.color;
  texture                = _draw.texture;
  connectivity_triangles = _draw.connectivity_triangles;
  connectivity_quads     = _draw.connectivity_quads;
  size_other_polygons    = _draw.size_other_polygons;
  connectivity_polygons  = _draw.connectivity_polygons;
}
Drawable_mesh::~Drawable_mesh()
{destroy();}

int Drawable_mesh::destroy()
{
  point_set.resize(0);
  normal.resize(0);
  color.resize(0);
  texture.resize(0);

  connectivity_triangles.resize(0);
  connectivity_quads.resize(0);
  size_other_polygons.resize(0);
  connectivity_polygons.resize(0);

  texture_number=-1;

  return 0;
}

int Drawable_mesh::build_drawable(const Mesh_object& _mesh)
{

  destroy();

  /// save in per polygon per vertex form every vertices (vertices duplicated)

  int k_polygon=0;
  int N_polygon=_mesh.polygon_number();
  std::vector <int> current_polygon;
  int current_index=0;
  int current_size=0;
  int k_vertex=0;
  int k_dim=0;
  double t0=0,t1=0;
  double r=0,g=0,b=0,t=0;

  set_texture_number(_mesh.get_texture_number());

  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
      current_polygon = _mesh.get_index_of_polygon(k_polygon);
      current_size=current_polygon.size();

      for(k_vertex=0;k_vertex<current_size;k_vertex++)
	{

	  //***********//
	  //GEOMETRY   //
	  //***********//
	  for(k_dim=0;k_dim<3;k_dim++)
	    point_set.push_back(_mesh.get_vertex(k_polygon,k_vertex,k_dim)); //vertices



	  //***********//
	  //NORMAL     //
	  //***********//	  
	  if(_mesh.normal_type()==1)//per polygon
	    for(k_dim=0;k_dim<3;k_dim++)
	      normal.push_back(_mesh.get_polygon_normal(k_polygon)[k_dim]);
	  else if(_mesh.normal_type()==2)//per vertex
	    for(k_dim=0;k_dim<3;k_dim++)
	      normal.push_back(_mesh.get_vertex_normal(k_polygon,k_vertex)[k_dim]);
	  else if(_mesh.normal_type()==3)//per vertex and polygon
	    for(k_dim=0;k_dim<3;k_dim++)
	      normal.push_back(_mesh.get_polygon_and_vertex_normal(k_polygon,k_vertex)[k_dim]);
	  
	  //***********//
	  //TEXTURE    //
	  //***********//	  
	  if(_mesh.texture_type()!=-1)
	    if(_mesh.texture_type()==0)//per polygon and vertex
	      {
		_mesh.get_intermediate_texture(k_polygon,k_vertex,&t0,&t1);
		texture.push_back(t0);texture.push_back(1.0-t1);
	      }

	  //***********//
	  //COLOR      //
	  //***********//	  
	  if(_mesh.color_type()==1)//per vertex
	    for(k_dim=0;k_dim<4;k_dim++)
	      _mesh.get_per_vertex_color(k_polygon,k_vertex,&r,&g,&b,&t);
	  else if(_mesh.color_type()==2)//per vertex
	    for(k_dim=0;k_dim<4;k_dim++)
	      _mesh.get_per_polygon_color(k_polygon,&r,&g,&b,&t);
	  else if(_mesh.color_type()==3)//per vertex and polygon
	    for(k_dim=0;k_dim<4;k_dim++)
	      _mesh.get_per_polygon_and_vertex_color(k_polygon,k_vertex,&r,&g,&b,&t);
	  if(_mesh.color_type()>=1 && _mesh.color_type()<=3)
	    {
	      color.push_back(r);
	      color.push_back(g);
	      color.push_back(b);
	      color.push_back(t);
	    }
	}
      // triangle
      if(current_size==3)
	for(k_dim=0;k_dim<3;k_dim++)
	  connectivity_triangles.push_back(current_index+k_dim);
      //connectivity_triangles.insert(connectivity_triangles.end(),current_polygon.begin(),current_polygon.end());

      
      //quads
      if(current_size==4)
	for(k_dim=0;k_dim<4;k_dim++)
	  connectivity_quads.push_back(current_index+k_dim);
      //connectivity_quads.insert(connectivity_quads.end(),current_polygon.begin(),current_polygon.end());


      //polygons
      if(current_size>4)
	{
	  size_other_polygons.push_back(current_polygon.size());
	  for(k_dim=0;k_dim<current_size;k_dim++)
	    connectivity_polygons.push_back(current_index+k_dim);
	  //connectivity_polygons.insert(connectivity_polygons.begin(),current_polygon.begin(),current_polygon.end());
	}

      current_index += current_size;

    }

  return 0;
}


int Drawable_mesh::triangle_number() const
{return connectivity_triangles.size()/3;}

int Drawable_mesh::quad_number() const
{return connectivity_quads.size()/4;}

int Drawable_mesh::polygon_number() const
{return size_other_polygons.size();}

int Drawable_mesh::polygon_size(int k_polygon) const
{
  int N_polygon=polygon_number();
  if(k_polygon<0 || k_polygon>=N_polygon)
    {printf("ERROR in Drawable_mesh in polygon_size(k_polygon=%d), polygon number is %d\n",k_polygon,N_polygon);exit(-1);}

  return size_other_polygons[k_polygon];
 
}



double* Drawable_mesh::point_set_pointer() 
{return &point_set[0];}
double* Drawable_mesh::normal_pointer() 
{return &normal[0];}
double* Drawable_mesh::texture_pointer() 
{return &texture[0];}
double* Drawable_mesh::color_pointer() 
{return &color[0];}
int* Drawable_mesh::connectivity_triangles_pointer() 
{return &connectivity_triangles[0];}
int* Drawable_mesh::connectivity_quads_pointer() 
{return &connectivity_quads[0];}
int* Drawable_mesh::connectivity_polygons_size_pointer() 
{return &size_other_polygons[0];}
int* Drawable_mesh::connectivity_polygons_pointer() 
{return &connectivity_polygons[0];}


double Drawable_mesh::get_point_set(int k_vertex,int k_dim) const
{
  int N_vertex=point_set.size();
  if(3*k_vertex+k_dim>=N_vertex || k_vertex<0)
    {printf("Error in get_point_set(%d,%d) in Drawable_mesh, point_set.size()=%d\n", k_vertex,k_dim,N_vertex);exit(-1);}
  return point_set[3*k_vertex+k_dim];
}

double Drawable_mesh::get_connectivity_triangles(int k_triangle,int k_dim) const
{
  int N_triangle=connectivity_triangles.size();
  if(3*k_triangle+k_dim>=N_triangle || k_triangle<0)
    {printf("Error in get_connectivity_triangle(%d,%d) in Drawable_mesh, connectivity_triangles.size()=%d\n", k_triangle,k_dim,N_triangle);exit(-1);}
  return connectivity_triangles[3*k_triangle+k_dim];
}

double Drawable_mesh::get_connectivity_quads(int k_quad,int k_dim) const
{
  int N_quad=connectivity_quads.size();
  if(3*k_quad+k_dim>=N_quad || k_quad<0)
    {printf("Error in get_connectivity_quads(%d,%d) in Drawable_mesh, connectivity_quads.size()=%d\n", k_quad,k_dim,N_quad);exit(-1);}
  return connectivity_quads[3*k_quad+k_dim];
}

int Drawable_mesh::is_normal() const
{
  if(normal.size()>0 && normal.size()==point_set.size())
    return 1;
  return 0;
}

int Drawable_mesh::is_texture() const
{
  if(texture.size()>0 && texture.size()/2==point_set.size()/3)
    return 1;
  return 0;
}
int Drawable_mesh::is_color() const
{
  if(color.size()>0 && color.size()/4==point_set.size()/3)
    return 1;
  return 0;
}

int Drawable_mesh::set_texture_number(int _texture_number)
{texture_number=_texture_number;return 0;}
int Drawable_mesh::get_texture_number() const
{return texture_number;}
