//** Modified-Light version of OpenGL_drawer for TP**// 

#include <OpenGL_drawer.h>

OpenGL_drawer::OpenGL_drawer():draw_list()
{
  draw_list.resize(0);
}

OpenGL_drawer::~OpenGL_drawer()
{
  unsigned int k=0;
  for(k=0;k<draw_list.size();k++)
    {
      if(glIsList(draw_list[k]))
	{
	  glDeleteLists(draw_list[k],1);

	  #ifdef DEBUG
	  printf("[draw_list %d deleted]\n",k);
	  #endif 
	}
    }

}

int OpenGL_drawer::init_opengl(int argc,char **argv,const Navigator_tool& navigator)
{
  //init window
  glutInit (&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowPosition(0,0);
  glutInitWindowSize(navigator.get_X_screen(),navigator.get_Y_screen());
  glutCreateWindow("win_1");


  //OpenGL set
  glClearColor(1.0f,1.0f,1.0f,1.0f);
  glPointSize(5);
  glLineWidth(2);

  glEnable(GL_COLOR_MATERIAL);
  //glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
  glEnable(GL_DEPTH_TEST);
  glDepthMask(GL_TRUE);


  glFrontFace(GL_CCW);//clockwise CW or CCW
  glDisable(GL_CULL_FACE);
  //glEnable(GL_CULL_FACE);
  glDisable(GL_LIGHTING);
  glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_FOG);



  //light
  float light_ambient[4]={0.4f,0.4f,0.4f,1.0f};
  float light_diffuse[4]={0.4f,0.4f,0.4f,1.0f};
  float light_specular[4]={1.0f,1.0f,1.0f,1.0f};
  float light_position [4] = {2.0f,100.0f,100.0f,1.0f};

  float material_ambient [4] = {1.0f,1.0f,1.0f,1.0f};
  float material_diffuse [4] = {1.0f,1.0f,1.0f,1.0f};
  float material_specular [4] = {1.0f,1.0f,1.0f,1.0f};
  float material_exponent [1] = {60.0f};


  glShadeModel(GL_SMOOTH);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);

  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material_ambient);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material_diffuse);  
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material_specular);  
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, material_exponent);
  glEnable(GL_DEPTH_TEST);


  glEnable(GL_LINE_SMOOTH);
  glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);

  return 0;
}











int OpenGL_drawer::draw_mesh_slow(Mesh_object& mesh)
{

  int N_polygons=0;
  int k_polygon=0;
  int N_vertex=0;
  int k_vertex=0;

  





  glEnable(GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
  glColor4d(0.5,0.0,0.2,1.0);
  double r=0.0,g=0.0,b=0.0,t=0.0;
  double t0=0.0,t1=0.0;

  //texture
  if(mesh.texture_type()==0)
    {
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,mesh.get_texture_number());
      glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    }

  N_polygons=mesh.polygon_number();
  V_3D x;
  V_3D n;
  for(k_polygon=0;k_polygon<N_polygons;k_polygon++)
    {
      N_vertex=mesh.polygon_size(k_polygon);
      glBegin(GL_POLYGON);
      for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
	{


	  x = mesh.get_vertex_v3d(k_polygon,k_vertex);

	  //NORMAL
	  if(mesh.normal_type()==1)//per polygon
	    n = mesh.get_polygon_normal(k_polygon);
	  else if(mesh.normal_type()==2)//per vertex
	    n = mesh.get_vertex_normal(k_polygon,k_vertex);
	  else if(mesh.normal_type()==3)//per vertex and polygon
	    n = mesh.get_polygon_and_vertex_normal(k_polygon,k_vertex);

	  //COLOR
	  if(mesh.color_type()==-1) //no color
	    glColor4d(0.5,0.0,0.2,1.0);
	  else if(mesh.color_type()==1) //per vertex
	    {
	      mesh.get_per_vertex_color(k_polygon,k_vertex,&r,&g,&b,&t);
	      glColor4d(r,g,b,t);
	    }
	  else if(mesh.color_type()==2) //per polygon
	    {
	      mesh.get_per_polygon_color(k_polygon,&r,&g,&b,&t);
	      glColor4d(r,g,b,t);
	    }	    
	  else if(mesh.color_type()==3) //per polygon_and_vertex
	    {
	      mesh.get_per_polygon_and_vertex_color(k_polygon,k_vertex,&r,&g,&b,&t);
	      glColor4d(r,g,b,t);
	    }	  

	  //TEXTURE
	  if(mesh.texture_type()==0)
	    {
	      mesh.get_intermediate_texture(k_polygon,k_vertex,&t0,&t1);
	      glTexCoord2f(GLfloat(t0),GLfloat(1.0-t1));
	    }

	  glNormal3d(n[0],n[1],n[2]);
	  glVertex3d(x[0],x[1],x[2]);
	  
	}
      glEnd();      
    }
  
  glDisable(GL_BLEND);

  if(mesh.texture_type()==0)
    {glDisable(GL_TEXTURE_2D);}
  


      
  return 0;
}

int OpenGL_drawer::draw_mesh_grid(Mesh_object& mesh)
{
  int N_polygons=0;
  int k_polygon=0;
  int N_vertex=0;
  int k_vertex=0;



  glDisable(GL_LIGHTING);
  glEnable(GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
  //glColor4d(0.5,0.8,0.2,1.0);
  
  N_polygons=mesh.polygon_number();
  V_3D x;
  for(k_polygon=0;k_polygon<N_polygons;k_polygon++)
    {
      N_vertex=mesh.polygon_size(k_polygon);
      glBegin(GL_LINE_STRIP);
      for(k_vertex=0;k_vertex<=N_vertex;k_vertex++)
	{
	  x = mesh.get_vertex_v3d(k_polygon,(k_vertex%N_vertex));
	  glVertex3d(x[0],x[1],x[2]);
	}
      glEnd();      
    }
      
  glDisable(GL_BLEND);
  glEnable(GL_LIGHTING);


  return 0;
}

int OpenGL_drawer::draw_segment(const Segment& s) const
{
  V_3D x0=s.get_start(),x1=s.get_last();
  glBegin(GL_LINES);
  glVertex3d(x0[0],x0[1],x0[2]);
  glVertex3d(x1[0],x1[1],x1[2]);
  glEnd();

  return 0;
}

int OpenGL_drawer::draw_triangle(const Triangle& t) const
{
  V_3D normal = t.normal();
  glBegin(GL_TRIANGLES);
  glNormal3d(normal[0],normal[1],normal[2]);
  glVertex3d(t[0][0],t[0][1],t[0][2]);
  glVertex3d(t[1][0],t[1][1],t[1][2]);
  glVertex3d(t[2][0],t[2][1],t[2][2]);
  glEnd();

  return 0;
}

int OpenGL_drawer::draw_polygon(const PPolygon& p)const
{
  V_3D normal = p.normal();

  glNormal3d(normal[0],normal[1],normal[2]);

  glBegin(GL_POLYGON);
  for(int k=0;k<p.size();k++)
    glVertex3d(p[k][0],p[k][1],p[k][2]);
  glEnd();

  return 0;
}



GLuint OpenGL_drawer::load_texture(const char* filename)
{
  GLuint texture_number;
  Texture tex;

  tex.load_picture(filename);


  glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
  glGenTextures (1, &texture_number);
  glBindTexture (GL_TEXTURE_2D, texture_number);

  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);//GL_REPEAT
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  if(tex.get_depth()==4)
    gluBuild2DMipmaps (GL_TEXTURE_2D, tex.get_depth(), tex.get_Nx(), tex.get_Ny(), GL_RGBA,GL_UNSIGNED_BYTE, tex.get_data());
  if(tex.get_depth()==3)
    gluBuild2DMipmaps (GL_TEXTURE_2D, tex.get_depth(), tex.get_Nx(), tex.get_Ny(), GL_RGB,GL_UNSIGNED_BYTE, tex.get_data());

  return texture_number;
}

double OpenGL_drawer::draw_fps(const Navigator_tool& navigator) const
{
  glDisable(GL_DEPTH_TEST);

  double fps=navigator.get_fps();
  char strfps[15]="\0";

  glPushMatrix();
  glLoadIdentity();	
  glRasterPos3f(-0.52f,+1.485f,-1.3f);
  sprintf ( strfps, "%4.1f fps", fps );
  for (unsigned int i=0;i<strlen(strfps);i++) 
    glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,*(strfps+i));
  glPopMatrix();


  glEnable(GL_DEPTH_TEST);

  return fps;
}

GLint OpenGL_drawer::add_draw_list(const Mesh_object& _mesh)
{
  Drawable_mesh d_mesh;
  d_mesh.build_drawable(_mesh);
  return add_draw_list(d_mesh,_mesh.get_name());
}

GLint OpenGL_drawer::set_draw_list(int k_list,const Mesh_object& _mesh)
{
  Drawable_mesh d_mesh;
  d_mesh.build_drawable(_mesh);
  return set_draw_list(k_list,d_mesh,_mesh.get_name());
}
GLint OpenGL_drawer::set_draw_list(int k_list,Drawable_mesh& d_mesh,const string& name)
{
  int new_draw_list=-1;

  new_draw_list=glGenLists(1);

  if(k_list==-1)//add at the end
    {
      draw_list.push_back(new_draw_list);
      name_mesh.push_back(name);
    }
  else
    {
      if(k_list<-1)
	{printf("Error in set_draw_list(%d,mesh)\n",k_list);exit(-1);}
      if(k_list>=int(draw_list.size())) //add an extra draw list
	{
	  draw_list.resize(k_list+1);
	  draw_list.push_back(new_draw_list);

	  name_mesh.resize(k_list+1);
	  name_mesh.push_back(name);
	}
      else //set an other draw list
	{
	  if(glIsList(draw_list[k_list]))//if exists, then delete and redo
	    glDeleteLists(draw_list[k_list],1);
	  draw_list[k_list]=new_draw_list;
	  name_mesh[k_list]=name;
	}
    }



  //get pointers
  double *vertex=d_mesh.point_set_pointer();
  double *normal=d_mesh.normal_pointer();
  double *texture=d_mesh.texture_pointer();
  double *color=d_mesh.color_pointer();
  int *connectivity_triangle=d_mesh.connectivity_triangles_pointer();
  int *connectivity_quad=d_mesh.connectivity_quads_pointer();
  int *size_polygon=d_mesh.connectivity_polygons_size_pointer();
  int *connectivity_polygon=d_mesh.connectivity_polygons_pointer();
  int N_tri=d_mesh.triangle_number();
  int N_quad=d_mesh.quad_number();
  int N_polygon=d_mesh.polygon_number();

  int k_polygon=0;
  int current_access=0;
  int current_size=0;


  //START CALL LIST
  glNewList(new_draw_list,GL_COMPILE);
  glEnable(GL_LIGHTING);

  //Z FIGHTING
  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(1.0,1.0);

  //VERTEX
  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(3,GL_DOUBLE,0,vertex);

  //NORMAL
  if(d_mesh.is_normal()==1)
    {
      glEnableClientState(GL_NORMAL_ARRAY);
      glNormalPointer(GL_DOUBLE,0,normal);
    }

  //TEXTURE
  if(d_mesh.is_texture()==1)
    {
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,d_mesh.get_texture_number());
      glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glTexCoordPointer(2,GL_DOUBLE,0,texture);
    }

  //COLOR
  if(d_mesh.is_color()==1)
    {
      glEnableClientState(GL_COLOR_ARRAY);
      glColorPointer(4,GL_DOUBLE,0,color);

      glEnable(GL_BLEND);
      glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
    }

  //DRAW TRIANGLE
  glDrawElements(GL_TRIANGLES,3*N_tri,GL_UNSIGNED_INT,connectivity_triangle);
  //DRAW QUADS
  glDrawElements(GL_QUADS,4*N_quad,GL_UNSIGNED_INT,connectivity_quad);
  //DRAW POLYGONS
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
      current_size=size_polygon[k_polygon];
      glDrawElements(GL_POLYGON,current_size,GL_UNSIGNED_INT,connectivity_polygon+current_access);      
      current_access+=current_size;
    }
  
  //DISABLE
  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  glDisableClientState(GL_COLOR_ARRAY);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);
  glDisable(GL_BLEND);
  glDisable(GL_POLYGON_OFFSET_FILL);	  

  glEndList();



  return new_draw_list;
}
GLint OpenGL_drawer::add_draw_list(Drawable_mesh& d_mesh,const string& name)
{return set_draw_list(-1,d_mesh,name);}


int OpenGL_drawer::draw_in_list(int k_list) const
{
  int N_list=draw_list.size();
  if(k_list<0 || k_list>N_list)
    {printf("Error in OpenGL_drawer in draw_in_list(%d), size of draw_list = %d\n",k_list,N_list);exit(-1);}

  if(glIsList(draw_list[k_list])!=GL_TRUE)
    {printf("Invalid draw list in draw_in_list(%d)\n",k_list);return -1;}

  glCallList(draw_list[k_list]);
  return 0;
}


int OpenGL_drawer::draw_stored_list() const
{
  int k_list=0;
  int N_list=draw_list.size();

  for(k_list=0;k_list<N_list;k_list++)
    if(glIsList(draw_list[k_list]))
      glCallList(draw_list[k_list]);
  return 0;
}

GLint OpenGL_drawer::update_draw_list(const Mesh_object& _mesh)
{
  string name_to_find=_mesh.get_name();
  int k_name=0,N_name=name_mesh.size();
  for(k_name=0;k_name<N_name;k_name++)
    if(name_mesh[k_name]==name_to_find)
      return set_draw_list(k_name,_mesh);
  return set_draw_list(-1,_mesh);
}
